#!python
# Copyright (c) 2000-2006 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.
#
# Shane Caraveo
#
# Provide file status information
# runs background thread to periodicaly update status on a file
# has plugin capability to add new types of status retreivers, i.e.
# (CVS, Perforce, Subversion)
#
# the status service gets all file information from the
# file service, and updates it with scc information
#
# observer notifications:
#   file_status (url)
#   file_status_updated (None)
#
# when a file status changes, the service sends a notification file_status_update.
# any observer that listens for this will receive the url of the file that has
# been updated.  It must then call back to the service to get the status
# array.
#

from xpcom import components, nsError, ServerException, COMException
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC
from xpcom.server import WrapObject, UnwrapObject

import logging
import timeline
import uriparse
import sys, os, time, stat, string
import threading, copy
import koprocessutils

log = logging.getLogger('koFileStatusService')
#log.setLevel(logging.DEBUG)

def sortFileStatus(a, b):
    if a.isDirectory and not b.isDirectory:
        return 1
    if b.isDirectory and not a.isDirectory:
        return -1
    return cmp(a.URI,b.URI)

class KoFileStatusService:
    _com_interfaces_ = [components.interfaces.koIFileStatusService, components.interfaces.nsIObserver]
    _reg_clsid_ = "{20732408-43DA-4ca2-BC9F-B82437A3CB2B}"
    _reg_contractid_ = "@activestate.com/koFileStatusService;1"
    _reg_desc_ = "Komodo File Status Service"

    def __init__(self):
        timeline.enter('koFileStatusService.__init__')
        self._proxyMgr = components.classes["@mozilla.org/xpcomproxy;1"].\
            getService(components.interfaces.nsIProxyObjectManager)
        self._observerSvc = components.classes["@mozilla.org/observer-service;1"].\
            getService(components.interfaces.nsIObserverService)
        self._observerProxy = self._proxyMgr.getProxyForObject(None,
            components.interfaces.nsIObserverService, self._observerSvc,
            PROXY_ALWAYS | PROXY_SYNC)
        self._globalPrefs = components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService).prefs
        self._fileSvc = \
            components.classes["@activestate.com/koFileService;1"] \
            .getService(components.interfaces.koIFileService)
        
        # set up data cache
        self._urllist = {}
        self._statusCheckers = []
        self.shutdown = 0
        self._continue = 0
        self._background = 0
        self._diskChecker = None
        
        # set up the background thread
        self._thread = None
        self._tlock = threading.Lock()
        self._cv = threading.Condition()
        self._shutdown = threading.Condition()
        timeline.leave('koFileStatusService.__init__')

    def init(self):
        timeline.enter('koFileStatusService.init')

        self._observerSvc.addObserver(self, "file_changed",0)
        self._observerSvc.addObserver(self, "file_status_now",0)
        self._observerSvc.addObserver(self, "file_update_now",0)
        self._observerSvc.addObserver(self, 'xpcom-shutdown', 1)
        
        self._fileSvc.observerService.addObserver(self,'',0)
        
        self._diskChecker = KoDiskFileChecker()
        self._statusCheckers.append(self._diskChecker)

        self._statusCheckers.append(KoCVSFileChecker())
        self._statusCheckers.append(KoP4FileChecker())
        self._statusCheckers.append(KoSVNFileChecker())

        
        for checker in self._statusCheckers:
            self._globalPrefs.prefObserverService.addObserver(self,checker.enabledPrefName,0)
            self._globalPrefs.prefObserverService.addObserver(self,checker.backgroundPrefName,0)

        self._tlock.acquire()
        self._thread = threading.Thread(target = KoFileStatusService._process,
                                        name = "File Status Service",
                                        args = (self,))
        self._thread.start()
        self._tlock.release()
        timeline.leave('koFileStatusService.init')

    def _removeObserver(self, topic):
        try:
            self._observerSvc.removeObserver(self, topic)
        except:
            log.debug("Unable to remove observer %s"%topic)

    def unload(self):
        self.shutdown = 1;
        if self._thread.isAlive():
            self._cv.acquire()
            self._cv.notify()
            self._cv.release()
            self._shutdown.acquire()
            self._shutdown.wait(5)
            self._shutdown.release()

        for checker in self._statusCheckers:
            try:
                self._globalPrefs.prefObserverService.removeObserver(self,checker.enabledPrefName)
                self._globalPrefs.prefObserverService.removeObserver(self,checker.backgroundPrefName)
            except:
                # prefs shutdown already?
                log.debug("Unable to remove prefs observers")
    
        self._removeObserver("file_changed")
        self._removeObserver("file_status_now")
        self._removeObserver("file_update_now")

        try:
            self._fileSvc.observerService.removeObserver(self,'')
        except:
            # file service shutdown before us?
            log.debug("Unable to remove file service observers")
    
    def notifyError(self, error, detail):
        # XXX 'scc_status_message' is evil (see bug 20568). In addition, it
        #     is poorly named because the code is structured such that the
        #     non-scc checkers here could set an error.
        if error:
            #print "attempting to send notification"
            try:
                self._observerProxy.notifyObservers(error, 'scc_status_message', detail)
            except COMException, e:
                pass
        
    def notifyWarning(self, warningTitle, warningDetail):
        # XXX 'scc_status_message' is evil (see bug 20568).
        warning = components.classes["@activestate.com/koStatusMessage;1"]\
                .createInstance(components.interfaces.koIStatusMessage)
        warning.category = "scc_warning_message"
        warning.msg = warningTitle
        warning.timeout = 5000
        warning.highlight = 1
        #print "attempting to send scc warning notification"
        try:
            self._observerProxy.notifyObservers(warning, 'scc_status_message', warningDetail)
        except COMException, e:
            pass
        
    def _getFileStatusNow(self, URI, forceRefresh=False):
        url = self._fileSvc.getFileFromURI(URI)
        self._getFileStatusNowWithFile(url, forceRefresh)
    
    def _getFileStatusNowWithFile(self, url, forceRefresh=False):
        url = UnwrapObject(url)
        lasterror = None
        # force an update to the status
        urllist = self._fileSvc.getFilesInBaseURI(url.URI)
        urllist = [UnwrapObject(_url) for _url in urllist]
        for checker in self._statusCheckers:
            if not checker.isActive():
                continue
            if checker.executablePrefName:
                checker.setExecutable(self._globalPrefs.getStringPref(checker.executablePrefName))

            if checker.scc and (not url.isLocal or \
                (url.sccDirType and url.sccDirType != checker.scc) or\
                (url.sccType and url.sccType != checker.scc)):
                continue

            # invalidate all the url's if it's a directory
            checker.invalidateUrl(url,1)

            # Clear the paths checked from previous run
            checker.clearRunStatusHistory()

            # No warnings or errors yet
            checker.warning = None
            checker.error = None

            # only check for the primary url.  if it is a directory
            # getFileStatus will get status for all files in that
            # directory, but not recursivly
            checker.getFileStatus(url, not url.isDirectory,
                                  forceRefresh=forceRefresh)
            if checker.error and (not lasterror or checker.error.msg != lasterror.msg):
                lasterror = checker.error
                self.notifyError(checker.error,checker.error_detail)
                checker.error = None
            else:
                if checker.warning:
                    self.notifyWarning(checker.warning, checker.warning_detail)
                    checker.warning = None
                # reset the status indicators
                # we don't do error checking here, these are associated uri's
                for _url in urllist:
                    if _url != url:
                        checker.getFileStatus(_url, not _url.isDirectory,
                                              forceRefresh=forceRefresh)
                    #else:
                    #    print "Not running getFileStatus() twice for the same url"
        
        # send notifications of status updates
        for _url in urllist:
            try:
                self._observerSvc.notifyObservers(_url, 'file_status', _url.URI)
            except COMException, e:
                if e.errno == nsError.NS_ERROR_FAILURE:
                    # if noone is listening, we get an exception
                    pass
                else:
                    #print "caught unknown exception in notify"
                    raise
        try:
            self._observerSvc.notifyObservers(self, 'file_status_updated',None)
        except COMException, e:
            if e.errno == nsError.NS_ERROR_FAILURE:
                # if noone is listening, we get an exception
                pass
            else:
                #print "caught unknown exception in notify"
                raise

    def getStatusForFile(self, kofile, forceRefresh):
        #print "getStatusForFile: %r" % (kofile.URI)
        self._getFileStatusNowWithFile(kofile, forceRefresh)

    def getStatusForUri(self, URI, forceRefresh):
        #print "getStatusForUri: %r" % (URI)
        self._getFileStatusNow(URI, forceRefresh)

    def isRepository(self, URI):
        url = self._fileSvc.getFileFromURI(URI)
        if not url or not url.isLocal or not url.exists:
            return ''
        
        if url.sccType:
            return url.sccType
        if url.sccDirType:
            return url.sccDirType
            
        for checker in self._statusCheckers:
            if not checker.isActive() or not checker.scc:
                continue
            if checker.isRepository(url.path):
                return checker.scc

        return ''
    
    # nsIObserver interface
    def observe(self, subject, topic, data):
        log.debug("status observer received %s:%s"%(topic,data))
        if topic == 'xpcom-shutdown':
            log.debug("file status got xpcom-shutdown, unloading");
            self.unload()
            return

        if topic == self._globalPrefs.id:
            # url is actually the pref name that was changed
            for checker in self._statusCheckers:
                if data == checker.enabledPrefName or \
                   data == checker.backgroundPrefName:
                    enabled = self._globalPrefs.getBooleanPref(checker.enabledPrefName)
                    background = self._globalPrefs.getBooleanPref(checker.backgroundPrefName)
                    if enabled != checker.enabled or background != checker.background:
                        checker.enabled = enabled
                        checker.background = background
                        self._cv.acquire()
                        self._cv.notify()
                        self._cv.release()
                        break
            # we don't continue in this observer for pref changes
            return
        
        elif topic == "file_status_now":
            #_url.update = 0
            self._getFileStatusNow(data)
            return
        elif topic == "file_changed":
            _url = self._fileSvc.getFileFromURI(data)
            _url = UnwrapObject(_url)
            _url.update = 0
            _url.dofileupdate = 1

        self._cv.acquire()
        # notify is an immediate release, but we want the thread
        # to continue until there is no more activity, so we flag
        # it to do so
        self._continue = 1
        self._cv.notify()
        self._cv.release()

    def _process(self):
        #print "starting file status background thread"
        urllist = []
        lasterror = None
        while 1:
            try:
                updated = []
                
                # if we have more url's than we checked last
                # time, lets do them now so the user doesn't
                # have to wait 30 seconds
                if not self._continue:
                    self._cv.acquire()
                    self._cv.wait(30)
                    # if self._continue is set, this is not a timeout
                    self._background = not self._continue
                    self._cv.release()
                # reset flag, see Observer for more info
                self._continue = 0
                
                if self.shutdown:
                    self._shutdown.acquire()
                    self._shutdown.notify()
                    self._shutdown.release()
                    log.info("file status thread shutting down")
                    return
                
                
                #print "doing file status update"
                urllist = self._fileSvc.getAllFiles()
                urllist = [UnwrapObject(_url) for _url in urllist]
                
                start = time.time()
                
                # invalidate old status information
                # we do this seperately so that the scc modules can
                # catch full directories of urls
                for checker in self._statusCheckers:
                    #log.debug("try %s"%checker.scc)
                    # do we need to run this checker?
                    if not checker.isActive():
                        #log.debug("Skipping %s, not active"%checker.scc)
                        continue
                    if checker.executablePrefName:
                        checker.setExecutable(self._globalPrefs.getStringPref(checker.executablePrefName))
                    checker.background = self._globalPrefs.getBooleanPref(checker.backgroundPrefName)
                    checker.recursive = checker.recursivePrefName and self._globalPrefs.getBooleanPref(checker.recursivePrefName)
                    checker.duration = self._globalPrefs.getLongPref(checker.durationPrefName)
                    if self._background and not checker.background:
                        #log.debug("Skipping %s, no background checking"%checker.scc)
                        continue
                    
                    # We start off with nothing already checked for this checker
                    checker.clearRunStatusHistory()

                    urls = []
                    # invalidate paths first so they can be recached
                    # build a list of url's that we can actually update
                    for url in urllist:
                        if url.isLocal and (url.update == 0 or \
                           (checker.background and checker.invalidatePath(url))):
                            # if this file is marked to have scc, and it's scc does
                            # not match the checker, then skip passing the file to the checker
                            if not checker.scc or (checker.scc and \
                                ((url.sccType and (url.sccType == checker.scc)) or \
                                (not url.sccDirType or (url.sccDirType == checker.scc)))):
                                urls.append(url)
                            #else:
                            #    log.debug('skipping %s, type [%s] dirType [%s]'%(url.URI, url.sccType, url.sccDirType))

                    urls.sort(sortFileStatus)
                    for url in urls:
                        # if we're in process, but a shutdown is requested, do it
                        if self.shutdown:
                            self._shutdown.acquire()
                            self._shutdown.notify()
                            self._shutdown.release()
                            #log.debug("status thread shutting down")
                            log.info("file status thread shutting down")
                            return
                        
                        if not checker.enabled:
                            #log.debug("Skipping %s, not enabled"%checker.scc)
                            break
                        
                        # if we're in a big long list (large project)
                        # then it's possible for the url to have been
                        # removed before we get here
                        #if url.remove: continue
                        
                        # check the status of the url
                        done = 0
                        #print "url time is %d, now is %d" %(self._urllist[url].update, time.time())
                        
                        try:
                            #log.debug("%s checking %s %s" %(checker.scc,url.sccType,url.URI))
                            oldscc = copy.deepcopy(url.scc)
                            oldstat = copy.deepcopy(url.stats)
                            
                            if url.dofileupdate:
                                checker.invalidateUrl(url,force=1)
                                
                            if not checker.getFileStatus(url, url.dofileupdate,
                                                         forceRefresh=False):
                                if checker.error and (not lasterror or checker.error.msg != lasterror.msg):
                                    lasterror = checker.error
                                    self.notifyError(checker.error,checker.error_detail)
                                    checker.error = None
                                #log.debug("Check Failed %s %s"%(checker.scc,checker.error))
                                continue
                            # this avoids making the notification below if the status has
                            # not changed.  The notification interupts and updates the UI
                            # taking an enormous amount of cpu usage
                            if oldscc == url.scc and \
                                oldstat == url.stats: continue
                            updated.append(url)
    
                            try:
                                if len(updated) % 10 == 0:
                                    tmpurllist = [u.URI for u in updated[-10:]]
                                    self._observerProxy.notifyObservers(None, 'file_status',
                                                                        string.join(tmpurllist, '\n'))
                            except COMException, e:
                                if e.errno == nsError.NS_ERROR_FAILURE:
                                    # noone is listening, we get an exception
                                    pass
                        except Exception, e:
                            log.exception("Unknown Exception %s" % repr(e))
                        
                    # on every checker we notify, since a particular checker
                    # can take some time
                    if updated:
                        try:
                            left = len(updated) % 10
                            if left > 0:
                                tmpurllist = [u.URI for u in updated[left * -1:]]
                                self._observerProxy.notifyObservers(None, 'file_status',
                                                                    string.join(tmpurllist, '\n'))
                        except COMException, e:
                            if e.errno == nsError.NS_ERROR_FAILURE:
                                # noone is listening, we get an exception
                                pass
                        try:
                            # this call is proxied to the UI thread.  The trees are invalidated
                            # in the observer for this call
                            self._observerProxy.notifyObservers(self, 'file_status_updated',None)
                        except COMException, e:
                            if e.errno == nsError.NS_ERROR_FAILURE:
                                # noone is listening, we get an exception
                                pass
                
                lasterror = None
                
                for url in urls:
                    url.dofileupdate = 0
                        
                end =  time.time()
                for url in updated:
                    url.update = end

                for url in urllist:
                    if url.update == 0:
                        url.update = end
            except:
                # we catch any exception so we can clear our thread variable
                # this allows us to restart the thread again later
                # we re raise the exception
                import traceback
                errmsg = ''.join(traceback.format_exception(*sys.exc_info()))
                log.error('KoFileStatusService thread exception %s' % errmsg)

        log.info("file status thread shutting down")

class KoFileCheckerBase:
    def __init__(self):
        self.type = None
        self.scc = None
        self.duration = 15
        self.durationPrefName = None
        self.enabledPrefName = None
        self.enabled = 0
        self.backgroundPrefName = None
        self.background = 0
        self.recursivePrefName = None
        self.recursive = 0
        self.error = None
        self.error_detail = None
        self.warning = None
        self.warning_detail = None
        self.executable = None
        self.executablePrefName = None
        # _runStatusHistory is what has been processed already by this checker.
        # The paths that have been checked will be placed into this dictionary,
        # if the path has been already checked, then there is no need to check
        # it again. This should be cleared every background run by the file
        # status checker.
        self._runStatusHistory = {}
        self._globalPrefs = components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService).prefs
    
    def setExecutable(self, executable):
        self.executable = executable
        
    def setError(self, msg='', detail=''):
        self.error = None
        if not msg: return
        
        self.error = components.classes["@activestate.com/koStatusMessage;1"]\
                .createInstance(components.interfaces.koIStatusMessage)
        self.error.category = "scc_error_message"
        self.error.msg = msg
        self.error.timeout = 5000
        self.error.highlight = 1
        self.error_detail = str(detail)
    
    def setWarning(self, msg='', detail=''):
        self.warning = msg
        self.warning_detail = detail
    
    def isRepository(self, path):
        return 0
    
    def isActive(self):
        self.enabled = self._globalPrefs.getBooleanPref(self.enabledPrefName)
        return self.enabled

    def invalidatePath(self, url,force=0):
        return 0

    def invalidateUrl(self, url,force=0):
        return 0

    def getFileStatus(self, url, dofile=1, forceRefresh=False):
        return None

    def clearRunStatusHistory(self):
        self._runStatusHistory = {}

class KoDiskFileChecker(KoFileCheckerBase):
    def __init__(self):
        timeline.enter('KoDiskFileChecker.__init__')
        KoFileCheckerBase.__init__(self)
        self.type = 'disk'
        self.durationPrefName = 'diskBackgroundMinutes'
        self.enabledPrefName = 'diskStatusEnabled'
        self.backgroundPrefName = 'diskBackgroundCheck'
        timeline.leave('KoDiskFileChecker.__init__')

    def invalidatePath(self, url,force=0):
        return force or url.update < time.time() - (self.duration * 60)

    def invalidateUrl(self, url,force=0):
        return force or url.update < time.time() - (self.duration * 60)

    def getFileStatus(self, url, dofile=1, forceRefresh=False):
        self.setError()
        changed = url.hasChanged
        # we mark dofileupdate to force
        # an scc update to occure
        if changed:
            url.dofileupdate = 1
        return changed


import p4lib
class KoP4FileChecker(KoFileCheckerBase):
    def __init__(self):
        timeline.enter('KoP4FileChecker.__init__')
        KoFileCheckerBase.__init__(self)
        self._have = {}
        self._not = {}
        #XXX rename _paths to _directories
        self._paths = {}
        self.p4 = p4lib.P4()
        self.type = 'scc'
        self.scc = 'p4'
        self.durationPrefName = 'p4BackgroundMinutes'
        self.enabledPrefName = 'p4Enabled'
        self.backgroundPrefName = 'p4BackgroundCheck'
        self.recursivePrefName = 'p4Recursive'
        self._p4Svc = components.classes["@activestate.com/koP4;1"].\
            getService(components.interfaces.koIP4)
        self.setExecutable(self._p4Svc.executable)
        self.executablePrefName = 'p4Executable'
        timeline.leave('KoP4FileChecker.__init__')

    def setExecutable(self, executable):
        if sys.platform.startswith("win"):
            self.executable = "p4.exe"
        else:
            self.executable = "p4"
        if executable and os.path.exists(executable):
            self.executable = executable
        
    def notifyError(self, e):
        if str(e).find("is not under client's root") != -1:
            return
        log.debug(e)
        self.setError('P4 Status Error, see SCC Output pane for details',e)

    # we don't cache the where information into _path here, since that
    # would prevent getFileStatus from running correctly (ie retrieving
    # files/have/open data)
    def isRepository(self, path):
        try:
            npath = os.path.normcase(path)
            if npath in self._not: return 0
            if npath in self._paths: return 1
            try:
                self.p4.p4 = self.executable
                basename = path
                if not os.path.isdir(path):
                    basename = os.path.dirname(path)
                if self.p4.where(path,cwd=basename): return 1
            except p4lib.P4LibError, e:
                self.notifyError(e)
                # XXX what is right? should we _not it?
                return 0
            self._not[npath] = time.time()
        except OSError, e:
            # remotely working I sometimes get a OSError exception out of
            # process.py read, which I assume is some kind of timeout, so
            # we'll ignore it and next time around try again
            pass
        return 0
    
    def isActive(self):
        self.enabled = self._globalPrefs.getBooleanPref(self.enabledPrefName)
        return self._p4Svc.isFunctional and self.enabled

    # returns true if the path is invalidated
    def invalidatePath(self, url, force=0):
        if not url.path:
            return 0
        if url.isDirectory:
            return self._invalidate(url.path, force)
        else:
            return self._invalidate(url.dirName, force)
    
    # returns true if the path is invalidated    
    def invalidateUrl(self, url, force=0):
        if not url.path:
            return 0
        return self._invalidate(url.path, force)
        
    def _invalidate(self, path, force):
        npath = os.path.normcase(path)
        if npath in self._paths:
            # this will force a refetch of the entire path
            # if the last stat on the path is older than
            # the duration
            lastcheck = self._paths[npath]
            if force or lastcheck < time.time() - self.duration *60:
                self._paths[npath] = 0
            else:
                return 0
        elif npath in self._not:
            lastcheck = self._not[npath]
            if force or lastcheck < time.time() - self.duration *60:
                self._not[npath] = 0
            else:
                return 0
        elif npath in self._have:
            # this will force a refetch of just the one file
            del self._have[npath]
        return 1
    
    def getFileStatus(self, url, dofile=0, forceRefresh=False):
        # initialize all items.  This helps to clear the item properties
        # when it is removed from scc
        _sccInfo = {}
        
        # we use nbasepath and npath for keys, we use
        # path and basepath for sending to P4 commands
        # this is because P4 is case sensitive
        basepath = path = url.path
        if not path:
            url.clearscc()
            return None
        nbasepath = npath = os.path.normcase(path)
        #print "CHECKING p4 for ",basepath
        if not url.isDirectory:
            basepath = url.dirName
            nbasepath = os.path.normcase(basepath)

        try:
            if nbasepath in self._not:
                url.clearscc()
                return None
            
            self.p4.p4 = self.executable
            
            if (nbasepath not in self._paths or \
                self._paths[nbasepath] == 0) and \
                os.path.isdir(basepath):
                if nbasepath in self._runStatusHistory:
                    # We just checked this path, no need to do it again
                    #print "Already checked! nbasepath:'%s' npath:'%s'" % (nbasepath, npath)
                    return None
                # execute p4 on the path
                if self.recursive:
                    subpath = "/..."
                else:
                    subpath = "/*"
                    
                #print "p4 checking ",basepath + subpath
                p4fstat = self.p4.fstat(basepath + subpath,
                                      cwd=basepath)
                self._runStatusHistory[nbasepath] = 1

                for file in p4fstat:
                    filekey = os.path.normcase(file['path'])
                    path = os.path.normcase(os.path.dirname(filekey))
                    self._paths[path]=time.time()
                    file['ondisk'] = os.path.exists(filekey);
                    self._have[filekey] = file
                
                if p4fstat:
                    self._paths[nbasepath]=time.time()

            if dofile and npath not in self._have and not os.path.isdir(npath):
                #print "checking p4 for ",path
                # file not in cache, get it now and stuff it into the cache
                p4fstat = self.p4.fstat(path,cwd=basepath)
                self._runStatusHistory[path] = 1
                for file in p4fstat:
                    filekey = os.path.normcase(file['path'])
                    path = os.path.normcase(os.path.dirname(filekey))
                    self._paths[path]=time.time()
                    file['ondisk'] = os.path.exists(filekey);
                    self._have[filekey] = file
                        
        
        except p4lib.P4LibError, e:
            self.notifyError(e)
            if nbasepath in self._paths:
                # this ensures that a path that IS in a repository, but contains
                # no files will be marked correctly
                #log.debug("no status, but is a p4 repository %s", nbasepath)
                _sccInfo['sccDirType']=self.scc
                url.scc = _sccInfo
            return None

        if not self._have.has_key(npath):
            # still don't have it, not a perforce file
            if nbasepath in self._paths:
                # this ensures that a path that IS in a repository, but contains
                # no files will be marked correctly
                #log.debug("no status, but is a p4 repository %s", nbasepath)
                _sccInfo['sccDirType']=self.scc
                url.scc = _sccInfo
            return None
        
        # get p4 info for the path, not the file
        
        have = self._have[npath]
        
        _sccInfo['sccType'] =  self.scc
        _sccInfo['sccDirType'] = self.scc
        _sccInfo['sccHaveOnDisk']= have['ondisk']
        _sccInfo['sccLocalRevision'] = str(have['haveRev'])
        _sccInfo['sccDepotRevision'] = str(have['headRev'])
        _sccInfo['sccNeedSync'] = int(_sccInfo['sccLocalRevision'] != _sccInfo['sccDepotRevision'])
        _sccInfo['sccConflict'] = '0'
        
        # necessary for status icons
        if _sccInfo['sccNeedSync']:
            _sccInfo['sccStatus'] = 'out of date'
            _sccInfo['sccSync'] = 1
        else:
            _sccInfo['sccStatus'] = 'up to date'
            _sccInfo['sccOk'] = 1

        _sccInfo['sccAction'] = have['action']
        _sccInfo['sccChange'] = have['change']
        
        url.scc = _sccInfo
        return 1


import cvslib
class KoCVSFileChecker(KoFileCheckerBase):
    def __init__(self):
        timeline.enter('KoCVSFileChecker.__init__')
        KoFileCheckerBase.__init__(self)
        self._files = {}
        # XXX change _paths to _directories
        self._paths = {}
        self._not = {}
        self._cvsSvc = components.classes["@activestate.com/koCVS;1"].\
            getService(components.interfaces.koICVS)
        self.cvs = cvslib.CVS()
        self.type = 'scc'
        self.scc = 'cvs'
        self.executable = self._cvsSvc.executable
        self.executablePrefName = 'cvsExecutable'
        self.duration = 15
        self.durationPrefName = 'cvsBackgroundMinutes'
        self.enabledPrefName = 'cvsEnabled'
        self.backgroundPrefName = 'cvsBackgroundCheck'
        self.recursivePrefName = 'cvsRecursive'
        timeline.leave('KoCVSFileChecker.__init__')

    def setExecutable(self, executable):
        if executable:
            self.executable = executable
        else:
            self.executable = 'cvs'

    # we don't cache the where information into _path here, since that
    # would prevent getFileStatus from running correctly (ie retrieving
    # files/have/open data)
    def isRepository(self, path):
        basepath = os.path.dirname(path)
        npath = os.path.normcase(basepath)
        if npath in self._not: return 0
        if npath in self._paths: return 1
        cvsRepositoryFile = os.path.join(basepath,'CVS','Repository')
        if os.path.exists(cvsRepositoryFile): return 1
        self._not[npath]=time.time()
        return 0
        
    def isActive(self):
        self.enabled = self._globalPrefs.getBooleanPref(self.enabledPrefName)
        return self._cvsSvc.isFunctional and self.enabled

    def invalidatePath(self, url, force=0):
        if not url.path:
            return 0
        if url.isDirectory:
            return self._invalidate(url.path, force)
        else:
            return self._invalidate(url.dirName, force)
        
    def invalidateUrl(self, url, force=0):
        if not url.path:
            return 0
        return self._invalidate(url.path, force)
        
    def _invalidate(self, path, force):
        #print "invalidating ",path
        npath = os.path.normcase(path)
        if npath in self._paths:
            # this will force a refetch of the entire path
            # if the last stat on the path is older than
            # the duration
            lastcheck = self._paths[npath]
            if force or lastcheck < time.time() - self.duration *60:
                self._paths[npath] = 0
            else:
                return 0
        elif npath in self._not:
            lastcheck = self._not[npath]
            if force or lastcheck < time.time() - self.duration *60:
                self._not[npath] = 0
            else:
                return 0
        elif npath in self._files:
            del self._files[npath]
        return 1

    def _raiseCVSError(self, msg):
        # since we have no way to know if this is
        # informational or not, we have to parse for
        # what we want to know
        if msg.find('authorization failed') > -1:
            raise cvslib.CVSLibError(msg)
    
    def getFileStatus(self, url, dofile=0, forceRefresh=False):
        # initialize all items.  This helps to clear the item properties
        # when it is removed from scc
        _sccInfo = {}
        
        #print "getFileStatus CVS %d %s"%(dofile,url.url)
        basepath = path = url.path
        if not path:
            url.clearscc()
            return None
        nbasepath = npath = os.path.normcase(path)
        
        if not url.isDirectory:
            basepath = url.dirName
            nbasepath = os.path.normcase(basepath)
        
        try:
            if nbasepath in self._not:
                url.clearscc()
                return None

            self.cvs._cvs = self.executable
            
            # cvs is recursive for status, so lets make sure that both the file
            # and the path do not exist before we stat a path
            if (nbasepath not in self._paths or \
                self._paths[nbasepath] == 0) and \
                os.path.isdir(basepath):
                
                # cvs status does not return status on paths, so we
                # have to build a status for the path :(
                cvsRepositoryFile = os.path.join(basepath,'CVS','Repository')
                if not os.path.exists(cvsRepositoryFile):
                    self._not[nbasepath]=time.time()
                    url.clearscc()
                    return None
                
                _sccInfo['sccDirType']=self.scc
                
                self._files[nbasepath] = {
                    'filename':'',
                    'status':'Up-to-date',
                    'rev':'1',
                    'revdate':'',
                    'rrev':'1',
                    'rpath':'',
                    'path':basepath,
                    'stickytag':'',
                    'stickydate':'',
                    'stickyoptions':'',
                    'edit': 0,
                    'ondisk':1
                    }
                
                #print "checking CVS for path ",basepath
                # execute cvs on the path
                # XXX figure out how to parse showTags later
                result = self.cvs.status(cwd=basepath,recursive=self.recursive)
                if 'stderr' in result:
                    self._raiseCVSError(result['stderr'])
                        
                #print repr(result)
                #print "got %d files" % (len(result['files']))
                for file in result['files']:
                    filekey = os.path.normcase(file['path'])
                    path = os.path.dirname(filekey)
                    self._paths[path]=time.time()
                    #print "have ",filekey
                    file['ondisk'] = int(os.path.exists(filekey))
                    self._files[filekey] = file
                
                self._paths[nbasepath]=time.time()

            if dofile and not self._files.has_key(npath) and \
               not os.path.isdir(path):
                #print "checking cvs for file ",path
                # file not in cache, get it now and stuff it into the cache
                filename = os.path.basename(path)
                result = self.cvs.status(files=filename,cwd=basepath)
                if 'stderr' in result:
                    self._raiseCVSError(result['stderr'])
                    
                #print repr(result)
                for file in result['files']:
                    filekey = os.path.normcase(file['path'])
                    #print "have ",filekey
                    file['ondisk'] = int(os.path.exists(filekey))
                    self._files[filekey] = file
                
        except cvslib.CVSLibError, e:
            #we get an exception here if the cvs library call errors out
            log.debug(e)
            self.setError('CVS Status Error, see SCC Output pane for details',e)
            url.clearscc()
            return None
        
        if not self._files.has_key(npath):
            # still don't have it, not a cvs file
            #print "cvs dont have ",npath
            # It is a CVS directory though, so that counts for something
            _sccInfo['sccDirType']=self.scc
            url.scc = _sccInfo
            return None
        
        # get cvs info for the path, not the file
        #print "setting info for ",npath
        file = self._files[npath]
        
        _sccInfo['sccType'] = self.scc
        _sccInfo['sccDirType'] = self.scc
        _sccInfo['sccHaveOnDisk']= file['ondisk']
        _sccInfo['sccLocalRevision'] = file['rev']
        _sccInfo['sccRevdate'] = file['revdate']
        _sccInfo['sccDepotRevision'] = file['rrev']
        _sccInfo['sccNeedSync'] = int(_sccInfo['sccLocalRevision'] != _sccInfo['sccDepotRevision'])
        _sccInfo['sccConflict'] = '0'
        _sccInfo['sccAction'] = ''
        _sccInfo['sccStatus'] = file['status']
        _sccInfo['sccChange'] = ''

        # this is independent of sync status
        if _sccInfo['sccStatus'].find('conflict') >= 0:
            _sccInfo['sccConflict'] = '1'

        # necessary for status icons
        if _sccInfo['sccNeedSync']:
            _sccInfo['sccSync'] = 1
        else:
            _sccInfo['sccOk'] = 1

        if (_sccInfo['sccLocalRevision'] and \
           (_sccInfo['sccLocalRevision'] == '0' or \
            _sccInfo['sccLocalRevision'][0] == '-')) or \
           _sccInfo['sccStatus'].find('Removed') >= 0:
            _sccInfo['sccAction'] = 'delete'
        elif (_sccInfo['sccRevdate'] and _sccInfo['sccRevdate'] == 'New File!') or \
            _sccInfo['sccStatus'].find('Added') >= 0:
            _sccInfo['sccAction'] = 'add'
        elif not _sccInfo['sccNeedSync'] and _sccInfo['sccStatus'] != 'Up-to-date':
            _sccInfo['sccAction'] = 'edit'
        elif file['edit']:
            _sccInfo['sccAction'] = 'edit'
            
        url.scc = _sccInfo
        return 1
    
    

import svnlib
class KoSVNFileChecker(KoFileCheckerBase):
    def __init__(self):
        timeline.enter('KoSVNFileChecker.__init__')
        KoFileCheckerBase.__init__(self)
        self._files = {}
        # XXX change _paths to _directories
        self._paths = {}
        self._not = {}
        self._svnSvc = components.classes["@activestate.com/koSVN;1"].\
            getService(components.interfaces.koISVN)
        self.svn = svnlib.SVN()
        self.type = 'scc'
        self.scc = 'svn'
        self.executablePrefName = 'svnExecutable'
        self.setExecutable(self._svnSvc.executable)
        self.duration = 15
        self.durationPrefName = 'svnBackgroundMinutes'
        self.enabledPrefName = 'svnEnabled'
        self.backgroundPrefName = 'svnBackgroundCheck'
        self.recursivePrefName = 'svnRecursive'
        timeline.leave('KoSVNFileChecker.__init__')

        # Subversion 1.3 can have "_svn" directories, so we check for this
        # http://subversion.tigris.org/svn_1.3_releasenotes.html#_svn-hack
        if sys.platform.startswith("win") and "SVN_ASP_DOT_NET_HACK" in os.environ:
            self._svnDirName = "_svn"
        else:
            self._svnDirName = ".svn"


    def setExecutable(self, executable):
        if executable:
            self.executable = executable
        else:
            self.executable = self._svnSvc.executable
        self.svn.svn = self.executable

    # we don't cache the where information into _path here, since that
    # would prevent getFileStatus from running correctly (ie retrieving
    # files/have/open data)
    def isRepository(self, path):
        basepath = os.path.dirname(path)
        npath = os.path.normcase(basepath)
        if npath in self._not: return 0
        if npath in self._paths: return 1
        if os.path.exists(os.path.join(basepath, self._svnDirName)): return 1
        self._not[npath]=time.time()
        return 0
        
    def isActive(self):
        self.enabled = self._globalPrefs.getBooleanPref(self.enabledPrefName)
        return self._svnSvc.isFunctional and self.enabled

    def invalidatePath(self, url, force=0):
        if not url.path:
            return 0
        if url.isDirectory:
            return self._invalidate(url.path, force)
        else:
            return self._invalidate(url.dirName, force)
        
    def invalidateUrl(self, url, force=0):
        if not url.path:
            return 0
        return self._invalidate(url.path, force)
        
    def _invalidate(self, path, force):
        #print "invalidating ",path
        npath = os.path.normcase(path)
        if npath in self._paths:
            # this will force a refetch of the entire path
            # if the last stat on the path is older than
            # the duration
            lastcheck = self._paths[npath]
            if force or lastcheck < time.time() - self.duration *60:
                self._paths[npath] = 0
            else:
                return 0
        elif npath in self._not:
            lastcheck = self._not[npath]
            if force or lastcheck < time.time() - self.duration *60:
                self._not[npath] = 0
            else:
                return 0
        elif npath in self._files:
            del self._files[npath]
        return 1

    def _raiseSVNError(self, msg):
        # since we have no way to know if this is
        # informational or not, we have to parse for
        # what we want to know
        #if msg.find('authorization failed') > -1:
            raise svnlib.SVNLibError(msg)

    def _stripEmptyLines(self, message):
        _result = []
        for line in message.splitlines(0):
            if line:
                _result.append(line)
        return '\n'.join(_result) + "\n"

    def getFileStatus(self, url, dofile=0, forceRefresh=False):
        # initialize all items.  This helps to clear the item properties
        # when it is removed from scc
        _sccInfo = {}
        # Svn can generate errors as well as results in the one call, we grab
        # all errors and put them in here.
        _errors = []
        
        #print "getFileStatus SVN %d %s"%(dofile,url.path)
        basepath = path = url.path
        if not path:
            url.clearscc()
            return None
        nbasepath = npath = os.path.normcase(path)
        
        if not url.isDirectory:
            basepath = url.dirName
            nbasepath = os.path.normcase(basepath)
        
        try:
            if nbasepath in self._not:
                url.clearscc()
                return None
            env = koprocessutils.getUserEnv()
            # svn is recursive for status, so lets make sure that both the file
            # and the path do not exist before we stat a path
            if (nbasepath not in self._paths or \
                self._paths[nbasepath] == 0) and \
                os.path.isdir(basepath):
                #print "path is a directory, get info: %r"%basepath
                # svn status does not return status on paths, so we
                # have to build a status for the path :(
                flist = os.listdir(basepath)
                if self._svnDirName not in flist:
                    self._not[nbasepath] = time.time()
                    url.clearscc()
                    return None
                del flist[flist.index(self._svnDirName)]
                
                _sccInfo['sccDirType']=self.scc
                
                self._files[nbasepath] = {
                    'filename':'',
                    'path':basepath,
                    'edit': 0,
                    'ondisk':0
                    }
                
                #print "checking SVN for path ",basepath
                # execute svn on the path
                # XXX figure out how to parse showTags later
                result, out = self.svn.statusEx(flist, cwd=basepath,
                                                non_recursive= not self.recursive,
                                                recursive=self.recursive,
                                                verbose=1,
                                                show_updates=forceRefresh,
                                                _raw=1, env=env)
                # The out parameter of svn statusEx is a tuple containing the output
                # from the info command and also the output from the status command.
                # out == (info_output, status_output)
                for out_type in out:
                    if out_type and 'stderr' in out_type and out_type['stderr']:
                        _errors.append(self._stripEmptyLines(out_type['stderr']))
                        
                #print repr(result)
                #print "got %d files" % (len(result))
                if result:
                    for fname, file in result.items():
                        filekey = os.path.normcase(os.path.join(basepath,file['Path']))
                        path = os.path.dirname(filekey)
                        self._paths[path]=time.time()
                        #print "have ",filekey
                        file['ondisk'] = int(os.path.exists(filekey))
                        self._files[filekey] = file
                elif _errors:
                    self._raiseSVNError("".join(_errors))

                self._paths[nbasepath]=time.time()

            if dofile and not self._files.has_key(npath) and \
               not os.path.isdir(path):
                #print "checking svn for file ",path
                # file not in cache, get it now and stuff it into the cache
                filename = os.path.basename(path)
                result,out = self.svn.statusEx(filename,
                                               recursive=self.recursive,
                                               non_recursive= not self.recursive,
                                               cwd=basepath,
                                               verbose=1,
                                               show_updates=forceRefresh,
                                               _raw=1, env=env)
                # The out parameter of svn statusEx is a tuple containing the output
                # from the info command and also the output from the status command.
                # out == (info_output, status_output)
                for out_type in out:
                    if out_type and 'stderr' in out_type and out_type['stderr']:
                        _errors.append(self._stripEmptyLines(out_type['stderr']))

                #print repr(result)
                if result:
                    for fname, file in result.items():
                        filekey = os.path.normcase(os.path.join(basepath,file['Path']))
                        #print "have ",filekey
                        file['ondisk'] = int(os.path.exists(filekey))
                        self._files[filekey] = file
                elif _errors:
                    self._raiseSVNError("".join(_errors))

        except svnlib.SVNLibError, e:
            # we get an exception here if the svn library call errors out
            log.debug(e)
            self.setError('SVN Status Error, see SCC Output pane for details',e)
            url.clearscc()
            return None

        if _errors:
            #print "SVN Errors: ", _errors
            # We have errors, but we also had results as well, so we make the
            # errors into a warning message. Make sure errors are one line messages.
            self.setWarning("SVN warning", "SVN Warning:\n" + "".join(_errors))

        if not self._files.has_key(npath):
            # still don't have it, not a svn file
            #print "svn dont have ",npath
            # It is a Subversion directory though, so that counts for something
            _sccInfo['sccDirType']=self.scc
            url.scc = _sccInfo
            return None
        
        # get svn info for the path, not the file
        #print "setting info for ",npath
        file = self._files[npath]
        #print file
        #{'History': ' ',
        # 'Last_Changed_Rev': '350',
        # 'Locked': ' ',
        # 'Name': 'asdf.txt',
        # 'URL': 'file',
        # 'Last Changed Date': '2004-12-07 14',
        # 'Switched': ' ',
        # 'Modified': ' ',
        # 'Repository UUID': 'a66e1112-37ea-0310-b8af-fdbad1884ccd',
        # 'reserved1': ' ',
        # 'Revision': '350',
        # 'Status': 'M',
        # 'reserved2': ' ',
        # 'Last_Changed_Author': 'shanec',
        # 'Schedule': 'normal',
        # 'Node Kind': 'file',
        # 'Sync': ' ',
        # 'Last Changed Author': 'shanec',
        # 'Checksum': 'a5890ace30a3e84d9118196c161aeec2',
        # 'Path': 'test1/asdf.txt',
        # 'Last Changed Rev': '350',
        # 'Text Last Updated': '2004-12-07 14'}
        
        _sccInfo['sccType'] = self.scc
        _sccInfo['sccDirType'] = self.scc
        _sccInfo['sccHaveOnDisk']= file['ondisk']
        _sccInfo['sccLocalRevision'] = '?'
        if 'Revision' in file:
            _sccInfo['sccLocalRevision'] = file['Revision']
            
        _sccInfo['sccRevdate'] = '?'
        if 'Last Changed Date' in file:
            _sccInfo['sccRevdate'] = file['Last Changed Date']
            
        _sccInfo['sccDepotRevision'] = '?'
        if 'Last Changed Rev' in file:
            _sccInfo['sccDepotRevision'] = file['Last Changed Rev']
        
        needSync = ('Sync' in file and file['Sync'] == '*')
        _sccInfo['sccNeedSync'] = '%d' % needSync
        _sccInfo['sccSync'] = _sccInfo['sccNeedSync']
        _sccInfo['sccOk'] = '%d' % (not needSync)
        _sccInfo['sccConflict'] = '0'
        _sccInfo['sccAction'] = ''
        _sccInfo['sccStatus'] = 'ok'
        _sccInfo['sccChange'] = ''

        if 'Status' in file:
            _sccInfo['sccAction'] = _sccInfo['sccStatus'] = svnlib.actionNames[file['Status']]
            _sccInfo['sccConflict'] = '%d' % int(_sccInfo['sccAction'] == 'conflict')
        # handle add/delete of a file so it can be commited
        if 'Schedule' in file and file['Schedule'] != 'normal':
            _sccInfo['sccAction'] = file['Schedule']
        #print _sccInfo
        url.scc = _sccInfo
        return 1

