#!/usr/bin/env python

""" Utility class used for the parsing of Javascript comments.

This uses the JavaSciptDoc style (JSDoc) "http://jsdoc.sourceforge.net/" for
allowing comments to specify specific information about the file structure.

These are the TAGS we use for JavaScript which is based upon what JSDoc
supplies. Note: YAHOO uses this style.

@base         (base class, inherits from)
@class        (class name)
@constructor  (constructor for the class)
@extends      (base class, extended by this class)
@final        (value is a constant)
@param        (a function parameter)
@private      (private variable or function)
@requires     (requires this other package in order to work)
@return       (type returned from function)
@returns      (  same as above  )
@static       (static class method)
@type         (variable type)

Example:

/**
 * Method for creating a slider
 *
 * @private
 * @param s {String} the name of the slider.
 * @param id {String} element id to place the silder within
 * @param {int} leftPadding is the size of the padding field on the left
 * @param rightPadding {int} optional field for setting the size of the padding
 * field on the right.
 * @return {Slider} a horizontal slider control
 */

Notes:
* comments and type information "{...}" are optional
* {} type field can be either the first or second position after the field.
* field comments can span multiple lines.
"""

class JSDocParameter:
    def __init__(self, paramname, paramtype=None, doc=None):
        self.paramname = paramname
        self.paramtype = paramtype
        self.doc = doc
    def __repr__(self):
        return "JSDocParameter: %r (%r) - %r" % (self.paramname, self.paramtype,
                                                 self.doc)

class JSDoc:
    A_CLASS = 1
    A_CONSTRUCTOR = 2
    A_PRIVATE = 4
    A_STATIC = 8
    A_CONSTANT = 16

    def __init__(self, comment=None):
        self._reset()
        if comment:
            # Full comment initially given
            #print "JSDoc comment: %r" % (comment)
            self.parse(comment)

    def __repr__(self):
        result = []
        if self.attributes:
            attrs = []
            if self.attributes & self.A_CLASS:
                attrs.append("class")
            elif self.attributes & self.A_CONSTRUCTOR:
                attrs.append("constructor")
            elif self.attributes & self.A_PRIVATE:
                attrs.append("private")
            elif self.attributes & self.A_STATIC:
                attrs.append("static")
            elif self.attributes & self.A_CONSTANT:
                attrs.append("constant")
            if len(attrs) > 0:
                result.append(" ".join(attrs))
        if self.baseclass:
            result.append("Baseclass:  %s" % (self.baseclass))
        for cp in self.params:
            result.append(str(cp))
        if self.type:
            result.append("Type:  %s" % (self.type))
        if self.returns:
            result.append("Returns:  %s" % (str(self.returns)))
        if self.doc:
            result.append("Doc:\n" + self.doc)
        return "JSDoc:\n  %s" % ("\n  ".join(result))

    def _reset(self):
        self.comment = None
        self.baseclass = None
        self.doc = None
        self.type = None
        self.attributes = 0
        # params is a list of JSDocParameter's
        self.params = []
        # returns is a JSDocParameter (does not have a paramname though)
        self.returns = None

    def _getTypeField(self, value):
        # Examples:
        #  'int'
        #  '{String}'
        #  'boolean|Object'
        #  'Array[](Number[])'
        if not value:
            return value

        # Only take first field if multiples are given
        pipePos = value.find('|')
        if pipePos > 0:
            value = value[:pipePos]

        value = value.strip()
        if value[-1] == "}":
            value = value[:-1]
            sp = value.split("{", 1)
            if len(sp) > 1:
                value = sp[1]
                sp = value.split(":", 1)
                if len(sp) > 1:
                    value = sp[1]

        # If a brace is in the value, it's an array
        bracePos = value.find("[")
        if bracePos >= 0:
            value = "Array"
        return value.strip()

    # Examples:
    #  "{Boolean}       true if the date is OOM"
    #  "el {HTMLElement} the element to animate"
    #  "{string}  sCategory  The log category for the message."
    def _getTypeFieldFromString(self, value):
        """Return tuple (type, rest of string)"""

        sp = value.split("{", 1)
        if len(sp) > 1:
            before = sp[0]
            value = sp[1]
            sp = value.split("}", 1)
            value = sp[0]
            if len(sp) > 1:
                after = sp[1]
                return (self._getTypeField(value), before + after)
        return (None, value)

    def _handle_base(self, value):
        self.baseclass = value

    # Same as base
    def _handle_extends(self, value):
        self._handle_base(value)

    def _handle_class(self, value):
        self.attributes |= self.A_CLASS

    def _handle_constructor(self, value):
        self.attributes |= self.A_CONSTRUCTOR

    def _handle_private(self, value):
        self.attributes |= self.A_PRIVATE

    def _handle_static(self, value):
        self.attributes |= self.A_STATIC

    def _handle_final(self, value):
        self.attributes |= self.A_CONSTANT

    def _handle_param(self, value):
        paramname = None
        paramtype = None
        doc = None
        sp = value.split(None, 2)
        for s in sp[:2]:
            if paramtype is None and s and s[0] == '{':
                # type information
                paramtype = self._getTypeField(s)
            elif paramname is None:
                paramname = s
        # Should have at least the paramname by now
        if paramname and paramtype:
            if len(sp) > 2:
                doc = sp[2]
            else:
                doc = None
        else:
            doc = " ".join(sp[1:3])
        cp = JSDocParameter(paramname, paramtype, doc)
        self.params.append(cp)

    def _handle_type(self, value):
        self.type = self._getTypeField(value)

    def _handle_return(self, value):
        returntype, doc = self._getTypeFieldFromString(value)
        if returntype:
            self.returns = JSDocParameter(None, returntype, doc)
    # Same as return
    def _handle_returns(self, value):
        return self._handle_return(value)

    def parse(self, comment):
        self._reset()
        self.comment = comment
        if not comment:
            return False
        doc = []
        lines = self.comment.splitlines(0)
        # Check to see if it's an actual javadoc
        docfield = None
        isJSDoc = False
        for line in lines:
            line = line.strip()
            #print "line: %r" % (line)
            if not isJSDoc:
                # Note: "*//**" style comes from the ciler using two comments
                if line in ("/**", "*//**"):
                    # It looks like a javadoc from here
                    isJSDoc = True
            else:
                # It's a javadoc, so parse up the fields
                if line == "*/":
                    isJSDoc = False
                elif line == "*//**":
                    self._reset()
                    self.comment = comment
                elif line == "*":
                    doc.append("")
                elif len(line) > 2 and line[:2] == "* ":
                    sp = line.split(None, 1)
                    #print sp
                    if len(sp) > 1 and sp[1][0] == '@':
                        # It's a javadoc field
                        #print sp
                        docfield = sp[1][1:]
                        sp = docfield.split(None, 1)
                        #print sp
                        #print "Tag: %r" % (sp[0])
                        handle_call = getattr(self, "_handle_%s" % (sp[0]), None)
                        if handle_call is not None:
                            if len(sp) == 1:
                                value = ""
                            else:
                                value = sp[1].strip()
                            handle_call(value)
                        # else: # We don't handle this param
                    elif not docfield: # This is likely the main doc string
                        doc.append(sp[1])
        self.doc = "\n".join(doc).rstrip()

    def isClass(self):
        return self.attributes & self.A_CLASS

    def isConstructor(self):
        return self.attributes & self.A_CONSTRUCTOR

    def isPrivate(self):
        return self.attributes & self.A_PRIVATE

    def isStatic(self):
        return self.attributes & self.A_STATIC

    def isConstant(self):
        return self.attributes & self.A_CONSTANT


############################################################
#                       Test code                          #
############################################################

def _test():
    sample_comment = """/**
 * Utility to set up the prototype, constructor and superclass properties to
 * support an inheritance strategy that can chain constructors and methods.
 *
 * @param {function} subclass   the object to modify
 * @param {function} superclass the object to inherit
 */
"""

    # Test the general usage of the class
    jd = JSDoc(sample_comment)
    assert(len(jd.params) == 2)
    assert(jd.params[0].paramname == "subclass")
    assert(jd.params[0].paramtype == "function")
    assert(jd.params[0].doc == "the object to modify")
    assert(jd.params[1].paramname == "superclass")
    assert(jd.params[1].paramtype == "function")
    assert(jd.params[1].doc == "the object to inherit")
    #print jd

    # Test specific internal functions of the class
    paramtype, doc = jd._getTypeFieldFromString("el {HTMLElement} the element to animate")
    assert(paramtype == "HTMLElement")
    assert(doc == "el  the element to animate")
    paramtype = jd._getTypeField("Array[](Number[])")
    assert(paramtype == "Array")
    paramtype = jd._getTypeField("Number|Array[])")
    assert(paramtype == "Number")

    jd._reset()
    jd._handle_param("{string}  sSource    The source of the the message (opt)")
    assert(len(jd.params) == 1 and \
           jd.params[0].paramname == "sSource" and \
           jd.params[0].paramtype == "string")

    jd._reset()
    jd._handle_param("oParent {Node} this node's parent node")
    assert(len(jd.params) == 1 and \
           jd.params[0].paramname == "oParent" and \
           jd.params[0].paramtype == "Node")

    jd._reset()
    jd._handle_returns("{array} Array of result objects")
    assert(jd.returns and \
           not jd.returns.paramname and \
           jd.returns.paramtype == "array")

    jd._reset()
    jd._handle_class("The superclass of all menu containers.")
    assert(jd.attributes & jd.A_CLASS)

    jd._reset()
    jd._handle_private("")
    assert(jd.attributes & jd.A_PRIVATE)

    jd._reset()
    jd._handle_static("")
    assert(jd.attributes & jd.A_STATIC)

    jd._reset()
    jd._handle_constructor("")
    assert(jd.attributes & jd.A_CONSTRUCTOR)

    jd._reset()
    jd._handle_base("YAHOO.widget.Menu")
    assert(jd.baseclass == "YAHOO.widget.Menu")
    jd._reset()
    jd._handle_extends("YAHOO.util.DragDrop")
    assert(jd.baseclass == "YAHOO.util.DragDrop")

    jd._reset()
    jd._handle_type("YAHOO.widget.MenuModuleItem")
    assert(jd.type == "YAHOO.widget.MenuModuleItem")
    jd._reset()
    jd._handle_type("{HTMLImageElement}")
    assert(jd.type == "HTMLImageElement")

# Main function
def main():
    _test()

# When run from command line
if __name__ == '__main__':
    main()
