#!/usr/bin/env python
# Copyright (c) 2000-2006 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.
#
#********************************************************************
# WARNING: Do not run this script directly. Run the main 
#          "install.sh" which will launch this script
#          properly.
#********************************************************************

"""
    install.sh - ActiveState Komodo "AS Package" install script

    Usage:
        install.sh [options...]

    General Options:
        -h, --help          print this help and exit
        -v, --verbose       verbose output

        -I, --install-dir <dir>     specify install directory
        -s, --suppress-shortcut     do NOT install desktop shortcut

    When called without arguments this script will interactively install
    Komodo. If the install dir is specified then Komodo will be
    installed without interaction.
"""

import sys
import os
from os.path import abspath, expanduser, normpath, join, dirname, basename
import shutil
import socket
import tempfile
import getopt
import logging
import pprint
import stat
import re
import subprocess
from glob import glob


class Error(Exception):
    pass



#---- globals

log = logging.getLogger("install")
gKomodoVer = "4.0"
gDefaultInstallDir = "/opt/Komodo-IDE-4.0"
gInstallLog = "koinstall.log"



#---- internal support routines

def _rmtreeOnError(rmFunction, filePath, excInfo):
    if excInfo[0] == OSError:
        # presuming because file is read-only
        os.chmod(filePath, 0777)
        rmFunction(filePath)

def _rmtree(dirname):
    import shutil
    shutil.rmtree(dirname, 0, _rmtreeOnError)

# Recipe: banner (1.0) in /Users/trentm/tm/recipes/cookbook
def _banner(text, ch='=', length=78):
    """Return a banner line centering the given text.

        "text" is the text to show in the banner. None can be given to have
            no text.
        "ch" (optional, default '=') is the banner line character (can
            also be a short string to repeat).
        "length" (optional, default 78) is the length of banner to make.

    Examples:
        >>> banner("Peggy Sue")
        '================================= Peggy Sue =================================='
        >>> banner("Peggy Sue", ch='-', length=50)
        '------------------- Peggy Sue --------------------'
        >>> banner("Pretty pretty pretty pretty Peggy Sue", length=40)
        'Pretty pretty pretty pretty Peggy Sue'
    """
    if text is None:
        return ch * length
    elif len(text) + 2 + len(ch)*2 > length:
        # Not enough space for even one line char (plus space) around text.
        return text
    else:
        remain = length - (len(text) + 2)
        prefix_len = remain / 2
        suffix_len = remain - prefix_len
        if len(ch) == 1:
            prefix = ch * prefix_len
            suffix = ch * suffix_len
        else:
            prefix = ch * (prefix_len/len(ch)) + ch[:prefix_len%len(ch)]
            suffix = ch * (suffix_len/len(ch)) + ch[:suffix_len%len(ch)]
        return prefix + ' ' + text + ' ' + suffix


if sys.platform.startswith("win"):
    def _getSystemDrive():
        try:
            return os.environ["SystemDrive"]
        except KeyError:
            raise Error("'SystemDrive' environment variable is not set")


def _getDefaultInstallDir():
    default = gDefaultInstallDir
    if sys.platform.startswith("win") and\
       default.lower().find("%systemdrive%") != -1:
        default = re.compile("%SystemDrive%", re.I).sub(_getSystemDrive(),
                                                        default)
    return default

def _askYesNo(question, default="yes"):
    """Ask the user the given question and their answer.
    
    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is one of "yes" or "no".
    """
    valid = {"yes":"yes",   "y":"yes",  "ye":"yes",
             "no":"no",     "n":"no"}
    if default == None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise Error("invalid default answer: '%s'" % default)

    while 1:
        sys.stdout.write(question + prompt)
        choice = raw_input().lower()
        #sys.stdout.write('\n')
        if default is not None and choice == '':
            return default
        elif choice in valid.keys():
            return valid[choice]
        else:
            sys.stdout.write("Please repond with 'yes' or 'no' "\
                             "(or 'y' or 'n').\n")


def _validateInstallDir(installDir):
    if os.path.exists(installDir) and not os.path.isdir(installDir):
        raise Error("cannot install to '%s': exists and is not a directory"
                    % installDir)


def _isBinary(filename):
    """Return true iff the given filename is binary.

    Raises an EnvironmentError if the file does not exist or cannot be
    accessed.
    """
    fin = open(filename, 'rb')
    try:
        CHUNKSIZE = 1024
        while 1:
            chunk = fin.read(CHUNKSIZE)
            if '\0' in chunk: # found null byte
                return 1
            if len(chunk) < CHUNKSIZE:
                break # done
    finally:
        fin.close()
    return 0


# Recipe: run (0.5.3) in C:\trentm\tm\recipes\cookbook
_RUN_DEFAULT_LOGSTREAM = ("RUN", "DEFAULT", "LOGSTREAM")
def __run_log(logstream, msg, *args, **kwargs):
    if not logstream:
        pass
    elif logstream is _RUN_DEFAULT_LOGSTREAM:
        try:
            log
        except NameError:
            pass
        else:
            if hasattr(log, "debug"):
                log.debug(msg, *args, **kwargs)
    else:
        logstream(msg, *args, **kwargs)

def _run(cmd, logstream=_RUN_DEFAULT_LOGSTREAM):
    """Run the given command.

        "cmd" is the command to run
        "logstream" is an optional logging stream on which to log the 
            command. If None, no logging is done. If unspecifed, this 
            looks for a Logger instance named 'log' and logs the command 
            on log.debug().

    Raises OSError is the command returns a non-zero exit status.
    """
    __run_log(logstream, "running '%s'", cmd)
    retval = os.system(cmd)
    if hasattr(os, "WEXITSTATUS"):
        status = os.WEXITSTATUS(retval)
    else:
        status = retval
    if status:
        #TODO: add std OSError attributes or pick more approp. exception
        raise OSError("error running '%s': %r" % (cmd, status))

def _run_in_dir(cmd, cwd, logstream=_RUN_DEFAULT_LOGSTREAM):
    """Run the given command in the given working directory.

        "cmd" is the command to run
        "cwd" is the directory in which the commmand is run.
        "logstream" is an optional logging stream on which to log the 
            command. If None, no logging is done. If unspecifed, this 
            looks for a Logger instance named 'log' and logs the command 
            on log.debug().

    Raises OSError is the command returns a non-zero exit status.
    """
    old_dir = os.getcwd()
    try:
        os.chdir(cwd)
        __run_log(logstream, "running '%s' in '%s'", cmd, cwd)
        _run(cmd, logstream=None)
    finally:
        os.chdir(old_dir)

def _run_and_capture(argv):
    log.debug("running '%s'", ' '.join(argv))
    p = subprocess.Popen(argv, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout = p.stdout.read()
    stderr = p.stderr.read()
    status = p.wait()
    return stdout, stderr, status


def _relocateFilesInDir((buildPrefix, installPrefix), dname, fnames):
    if len(installPrefix) > len(buildPrefix):
        raise Error("cannot properly relocate dir-dependent binaries "
                    "because install path is too long (%d chars): you "
                    "must install ActivePython to a path with fewer than "
                    "%d characters" % (len(installPrefix), len(buildPrefix)))
    for fname in fnames:
        fpath = os.path.join(dname, fname)
        if os.path.isdir(fpath):
            continue

        # Determine if this file needs to be relocated.
        fin = open(fpath, 'rb')
        try:
            content = fin.read()
        finally:
            fin.close()
        isBinary = _isBinary(fpath)
        if isBinary:
            this = os.path.join(buildPrefix, "lib")
            that = os.path.join(installPrefix, "lib")
        else:
            this = buildPrefix
            that = installPrefix
        if content.find(this) == -1:
            continue

        # Relocate this file.
        log.debug("relocate '%s'", fpath)
        perm = stat.S_IMODE(os.stat(fpath).st_mode)
        if isBinary:
            if sys.platform.startswith("aix"):
                # On AIX the lib path _list_ is stored as one string, rather
                # than just the one path. This means that the integrity of
                # the path list must be maintained by separating with ':'.
                # We also change the remainder to all x's to ensure it is
                # a bogus path.
                that = os.path.join(installPrefix, "lib")\
                       + ':' + "x"*(len(buildPrefix)-len(installPrefix)-1)
                log.debug("replace (length %d)\n\t%s\nwith (length %d)\n\t%s",
                          len(this), this, len(that), that)
                content = content.replace(this, that)
            else:
                # Replace 'this' with 'that' in a null-terminated string.
                # Make sure to properly correct for trailing content in the
                # same string because:
                # - on HP-UX sometimes a full path to the shared lib is stored:
                #      <this>/libtcl8.4.sl\0
                # - on AIX a path _list_ is stored:
                #      <this>:other/lib/paths\0
                #   NOTE: This *should* work on AIX, AFAICT, but it does
                #   *not*.
                pattern = re.compile("%s([^\0]*)\0" % re.escape(this))
                def c_string_replace(match, before=this, after=that):
                    lendiff = len(before) - len(after)
                    s = after + match.group(1) + ("\0"*lendiff) + "\0"
                    # Encode nulls as '0' instead of '\x00' so one can see
                    # the before and after strings line up.
                    log.debug("replace (length %d)\n\t%s\nwith (length %d)\n\t%s",
                              len(match.group(0)),
                              repr(match.group(0)).replace("\\x00", '0'),
                              len(s),
                              repr(s).replace("\\x00", '0'))
                    return s
                content = pattern.sub(c_string_replace, content)
        else:
            log.debug("replace (length %d)\n\t%s\nwith (length %d)\n\t%s",
                      len(this), this, len(that), that)
            content = content.replace(this, that)
        # Sometimes get the following error. Avoid it by removing file first.
        #   IOError: [Errno 26] Text file busy: '$fpath'
        os.remove(fpath)
        fout = open(fpath, 'wb')
        try:
            fout.write(content)
        finally:
            fout.close()
        os.chmod(fpath, perm) # restore permissions

def _install_desktop_shortcut(absInstallDir, suppressShortcut):
    """Install a desktop shortcut as appropriate.

        "absInstallDir" is the absolute path to which Komodo was
            just installed
        "suppressShortcut" is a boolean indicating if creation
            of the desktop shortcut is to be suppressed.

    - If we are install under $HOME then look for and install the
      desktop shortcut to $HOME/Desktop, if it exists.
    - Otherwise, attempt to add the .desktop shortcut to shared
      applications shortcuts dir.
    (See bug 32351 for details).
    """
    # Put together the shortcut content.
    content = """\
[Desktop Entry]
Encoding=UTF-8
Name=Komodo IDE 4.0
GenericName=IDE
Comment=A new dynamic in web development
Exec=%s/bin/komodo %%F
Icon=%s/share/icons/komodo48.png
Terminal=false
Type=Application
Categories=ActiveState;Application;Development;Debugger;IDE
""" % (absInstallDir, absInstallDir)
    shortcutName = "komodo-ide-4.0.desktop"
#    komodo-%s-%s.desktop" % ("ide", gKomodoVer)

    class ShortcutInstallError(Exception):
        pass

    try:
        if suppressShortcut:
            raise ShortcutInstallError("shortcut suppressed by user")

        # Determine if installing under HOME or to shared area.
        HOME = os.environ.get("HOME", None)
        if not HOME:
            raise ShortcutInstallError("no HOME environment variable")
        elif absInstallDir.startswith(HOME):
            shortcutDir = os.path.join(HOME, "Desktop")
        else:
            shortcutDir = "/usr/share/applications"
        shortcutPath = os.path.join(shortcutDir, shortcutName)

        # Attempt to write the Komodo shortcut.
        # (We DO overwrite an existing such shortcut.)
        if not os.path.exists(shortcutDir):
            raise ShortcutInstallError("'%s' does not exist" % shortcutDir)
        else:
            fout = open(shortcutPath, 'w')
            try:
                fout.write(content)
            finally:
                fout.close()
    except (EnvironmentError, ShortcutInstallError), ex:
        fallbackDir = os.path.join(absInstallDir, "share", "desktop")
        fallbackPath = os.path.join(fallbackDir, shortcutName)
        try:
            if not os.path.exists(fallbackDir):
                os.makedirs(fallbackDir)
            fout = open(fallbackPath, 'w')
            fout.write(content)
            fout.close()
        except EnvironmentError, ex2:
            log.warn("unexpected error creating fallback .desktop file "
                     "'%s': %s", fallbackPath, ex2)
        else:
            log.warn("did not install desktop shortcut: %s "
                     "(a Komodo .desktop file has been created in '%s' "
                     "that you may install manually)",
                     ex, fallbackPath)
    else:
        log.info("'Komodo %s' desktop shortcut created at '%s'",
                 gKomodoVer, shortcutPath)

def _reloc_komodo_stub(absInstallDir):
    """In <installdir>/bin there is a 'komodo' shell script stub to
    launch the real Komodo binary.

    That thing in its original state does not work if called via a
    symlink (which is very common). The right solution would be for it
    to effectively use realpath() to resolve the symlinks -- however, I
    don't have a realpath in Bash. So instead we patch it on install.
    """
    stubPath = os.path.join(absInstallDir, "bin", "komodo")
    log.debug("relocating stub: '%s'", stubPath)

    #XXX Sometimes have to chmod then file to be writable (good ol'
    #    Perforce).
    os.chmod(stubPath, 0755)

    fin = open(stubPath, 'r')
    try:
        content = fin.read()
    finally:
        fin.close()

    # Replace this line:
    #   INSTALLDIR=...
    # with the appropriate installdir.
    lines = content.splitlines(1)
    for i, line in enumerate(lines):
        if line.startswith("INSTALLDIR="):
            lines[i] = 'INSTALLDIR="%s"\n' % absInstallDir
    content = ''.join(lines)

    fout = open(stubPath, 'w')
    try:
        fout.write(content)
    finally:
        fout.close()

def _gen_so_paths(basedir):
    for dirpath, dirnames, filenames in os.walk(basedir):
        for filename in filenames:
            if filename.endswith(".so"):
                yield os.path.join(dirpath, filename)

def _selinux_prepare(absInstallDir):
    """If this is Linux and SELinux is installed and enabled,
    then we need to set the security context on the SciMoz plugin
    to allow shared object text relocation.

    See bug 43260 and bug 46275 for details.
    """
    if not sys.platform.startswith("linux"):
        return

    import selinuxlib
    selinux = selinuxlib.SELinux()
    
    if not selinux.is_installed():
        return

    for so_path in _gen_so_paths(absInstallDir):
        if not selinux.is_path_labeled(so_path):
            log.debug("%s: setting context just won't work, skipping", 
                      so_path)
            continue
        # Trying these covers RHEL (texrel_shlib_t), FC5 (textrel_shlib_t)
        # and CentOS (shlib_t).
        contexts_to_try = ["texrel_shlib_t", "textrel_shlib_t", "shlib_t"]
        context = selinux.context_from_path(so_path)
        if context is not None:
            context_to_try = context.split(':')[-1]
            if context_to_try not in contexts_to_try:
                contexts_to_try.append(context_to_try)
        for context_to_try in contexts_to_try:
            try:
                selinuxlib.chcon(so_path, context_to_try)
            except selinuxlib.SELinuxError, ex:
                pass
            else:
                break
        else:
            raise Error("could not set SELinux security context for "
                        "'%s': '%s' contexts failed"
                        % (so_path, "', '".join(contexts_to_try)))

def _install(installDir, userDataDir, suppressShortcut):
    normInstallDir = normpath(expanduser(installDir))
    absInstallDir = abspath(normInstallDir)
    pyInstallDir = join(absInstallDir, "lib", "python")
    mozInstallDir = join(absInstallDir, "lib", "mozilla")
    log.info("Installing ActiveState Komodo to '%s'...", normInstallDir)
    _validateInstallDir(absInstallDir)
    log.debug("redirect userDataDir to '%s'", userDataDir)
    log.debug("user environment: %s", pprint.pformat(dict(os.environ)))

    # copy the entire "Komodo" tree to the installDir
    if not os.path.exists(absInstallDir):
        os.makedirs(absInstallDir)
    import sh2
    sh2.cp(join(dirname(dirname(__file__)), "INSTALLDIR", "*"),
           dstdir=absInstallDir, preserve=True,
           recursive=True)

    # The copying above destroyed this symlink in ".../lib/mozilla":
    #   libpythonX.Y.so -> libpythonX.Y.so.1.0
    # Having two independent libpythonX.Y.so's results in the following
    # on Solaris:
    #   Fatal Python error: Interpreter not initialized (version mismatch?)
    # when doing PyXPCOM registration on startup. I don't know why on
    # Solaris and not on Linux. Restore the symlink for both.
    if sys.platform not in ("win32", "darwin"):
        libpythonXYso = glob(join(absInstallDir, "lib", "mozilla",
                                  "libpython?.?.so"))[0]
        _run('rm -f "%s"' % libpythonXYso)
        _run_in_dir("ln -s %s.1.0 %s"
                    % (basename(libpythonXYso), basename(libpythonXYso)),
                    dirname(libpythonXYso))

    log.debug("Preparing internal Python...")
    import activestate
    pyBuildPrefix = activestate.prefix
    log.debug("relocating siloed Python from '%s' to '%s'", pyBuildPrefix,
              pyInstallDir)
    os.path.walk(pyInstallDir, _relocateFilesInDir,
                 (pyBuildPrefix, pyInstallDir))

    log.debug("pre-compile .py files in siloed Python")
    pyLibDir = os.path.join(pyInstallDir, "lib",
                            "python%s.%s" % sys.version_info[:2])
    # Skip .pyc one, just .pyo's are necessary because we always run
    # Python in optimized mode.
    cmds = ['"%s" -W "ignore:hex/oct constants" -O "%s/compileall.py" '
            '-q -x test/ "%s"'
            % (sys.executable, pyLibDir, pyLibDir)]
    for cmd in cmds:
        log.debug("running '%s'" % cmd)
        retval = os.system(cmd)
        if retval:
            log.warn("error running '%s'" % cmd)

    # Configure siloed GTK2 libs (if any).
    log.debug("Configuring GTK2...")
    files = [os.path.join(mozInstallDir, "pango", "pangorc"),
             os.path.join(mozInstallDir, "pango", "pango.modules"),
             os.path.join(mozInstallDir, "gdk-pixbuf", "gdk-pixbuf.loaders")]
    token = "__PATH_TO_INSTALLED_KOMODO__"
    for file in files:
        if not os.path.exists(file):
            continue
        log.debug("fixing up '%s'", file)
        fin = open(file, 'r')
        try:
            content = fin.read()
        finally:
            fin.close()
        content = content.replace(token, absInstallDir)
        os.chmod(file, 0644)
        fout = open(file, 'w')
        try:
            fout.write(content)
        finally:
            fout.close()

    _install_desktop_shortcut(absInstallDir, suppressShortcut)
    _reloc_komodo_stub(absInstallDir)
    _selinux_prepare(absInstallDir)

    print """
%s
Komodo IDE 4.0 has been successfully installed to:
    %s
    
You might want to add 'komodo' to your PATH by adding the 
install dir to you PATH. Bash users can add the following
to their ~/.bashrc file:

    export PATH="%s/bin:$PATH"

Or you could create a symbolic link to 'komodo', e.g.:

    ln -s "%s/bin/komodo" /usr/local/bin/komodo

Documentation is available in Komodo or on the web here:
    http://aspn.activestate.com/ASPN/docs/Komodo

Please send us any feedback you might have or log bugs here:
    komodo-feedback@ActiveState.com
    http://bugs.activestate.com/Komodo/

Thank you for using Komodo.
%s
""" % (_banner(None), absInstallDir, absInstallDir, absInstallDir,
       _banner(None))



#---- main public functions

def interactiveInstall(suppressShortcut):
    default = _getDefaultInstallDir()
    sys.stdout.write("""\
Enter directory in which to install Komodo. Leave blank and
press 'Enter' to use the default [%s].
Install directory: """ % default)
    installDir = raw_input().strip()
    if not installDir:
        installDir = default

    norm = os.path.normpath(os.path.expanduser(installDir))
    if os.path.isdir(norm):
        sys.stdout.write("""
'%s' already exists. Installing over an existing
Komodo installation may have unexpected results. Are you
sure you would like to proceed with the installation?
""" % installDir)
        choice = _askYesNo("Proceed?", default="no")
        if choice == "yes":
            pass
        elif choice == "no":
            print "Aborting install."
            return
    elif os.path.exists(norm):
        raise Error("'%s' exists and is not a directory" % installDir)

    print
    install(installDir, suppressShortcut)

def install(installDir, suppressShortcut):
    # Redirect the "user data dir" to a temp location to avoid
    # the problem described in bug 32270 ("sudo ./install.sh" results in
    # interfering root-owned stuff in ~/.komodo).
    tempDir = os.path.join(tempfile.gettempdir(), "koinstall.%s" % os.getpid())
    try:
        return _install(installDir, tempDir, suppressShortcut)
    finally:
        if os.path.exists(tempDir):
            log.debug("removing temp user data dir: '%s'", tempDir)
            _rmtree(tempDir)



#---- mainline

# Recipe: pretty_logging (0.1.1) in C:\trentm\tm\recipes\cookbook
class _PerLevelFormatter(logging.Formatter):
    """Allow multiple format string -- depending on the log level.
    
    A "fmtFromLevel" optional arg is added to the constructor. It can be
    a dictionary mapping a log record level to a format string. The
    usual "fmt" argument acts as the default.
    """
    def __init__(self, fmt=None, datefmt=None, fmtFromLevel=None):
        logging.Formatter.__init__(self, fmt, datefmt)
        if fmtFromLevel is None:
            self.fmtFromLevel = {}
        else:
            self.fmtFromLevel = fmtFromLevel
    def format(self, record):
        record.levelname = record.levelname.lower()
        if record.levelno in self.fmtFromLevel:
            #XXX This is a non-threadsafe HACK. Really the base Formatter
            #    class should provide a hook accessor for the _fmt
            #    attribute. *Could* add a lock guard here (overkill?).
            _saved_fmt = self._fmt
            self._fmt = self.fmtFromLevel[record.levelno]
            try:
                return logging.Formatter.format(self, record)
            finally:
                self._fmt = _saved_fmt
        else:
            return logging.Formatter.format(self, record)


def main(argv):
    # Setup logging.
    root = logging.getLogger() # root logger

    hdlr = logging.FileHandler(gInstallLog, mode='w')
    formatter = logging.Formatter('%(asctime)s: %(levelname)s: %(name)s: %(message)s')
    hdlr.setFormatter(formatter)
    root.setLevel(logging.DEBUG)
    root.addHandler(hdlr)

    console = logging.StreamHandler()
    defaultFmt = "%(name)s: %(levelname)s: %(message)s"
    infoFmt = "%(name)s: %(message)s"
    fmtr = _PerLevelFormatter(fmt=defaultFmt,
                              fmtFromLevel={logging.INFO: infoFmt})
    console.setFormatter(fmtr)
    # Specifically don't log selinux messages on the console (bug 45757).
    onlyInstallFilter = logging.Filter("install")
    console.addFilter(onlyInstallFilter)
    console.setLevel(logging.INFO)
    root.addHandler(console)

    # Parse options.
    try:
        opts, args = getopt.getopt(argv[1:], "vhI:s",
            ["verbose", "help", "install-dir=", "suppress-shortcut"])
    except getopt.GetoptError, ex:
        log.error(str(ex))
        log.error("Try `install.sh --help'.")
        return 1
    installDir = None
    suppressShortcut = False
    for opt, optarg in opts:
        if opt in ("-h", "--help"):
            sys.stdout.write(__doc__)
            return
        elif opt in ("-v", "--verbose"):
            console.setLevel(logging.DEBUG)
        elif opt in ("-I", "--install-dir"):
            installDir = optarg
        elif opt in ("-s", "--suppress-shortcut"):
            suppressShortcut = True

    try:
        if installDir is None:
            interactiveInstall(suppressShortcut)
        else:
            install(installDir, suppressShortcut)
    except (EnvironmentError, Error), ex:
        log.error(str(ex))
        log.debug("exception info:", exc_info=True)
        return 1
    except KeyboardInterrupt:
        log.debug("user abort")
        pass

if __name__ == "__main__":
    __file__ == sys.argv[0]
    sys.exit( main(sys.argv) )

