/* Copyright (c) 2000-2006 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

// Utility methods to easily use the KoIInterpolationService from JavaScript.
// Basically you can use the Interpolate_interpolate() method to iterpolation
// special '%'-escape codes in a given list of strings. See
// koInterpolationService.py (or play around with "run commands") for an
// authoritative description of available escape codes.
//
// How to interpolate a couple of strings (mystringa, mystringb). You can do as
// many strings as you want.
//
//    var imystringa = null;
//    var imystringb = null;
//    try {
//        var istrings = Interpolate_interpolate(
//                          <ref-to-komodo.xul>,
//                          [mystringa, mystringb], // codes are not bracketed
//                          [mystringc, mystringd], // codes are bracketed
//                          <optional-query-title>);
//        imystringaForUse     = istrings[0];
//        imystringaForDisplay = istrings[1];
//        imystringbForUse     = istrings[2];
//        imystringbForDisplay = istrings[3];
//        imystringcForUse     = istrings[4];
//        imystringcForDisplay = istrings[5];
//        imystringdForUse     = istrings[6];
//        imystringcForDisplay = istrings[7];
//    } catch (ex) {
//        var errno = lastErrorSvc.getLastErrorCode();
//        if (errno == Components.results.NS_ERROR_ABORT) {
//            // Command was cancelled.
//        } else if (errno == Components.results.NS_ERROR_INVALID_ARG) {
//            var errmsg = lastErrorSvc.getLastErrorMessage();
//            alert("Could not interpolate:" + errmsg);
//        } else {
//            log.error(ex);
//            alert("There was an unexpected error: " + ex);
//        }
//    }
//


//---- globals

var _interpolate_log = getLoggingMgr().getLogger("interpolate");



//---- public interface

function Interpolate_isWordCharacter(ch)
{
    return /^\w/.test(ch);
}

function Interpolate_getWordUnderCursor(scimoz) {
    var start = scimoz.wordStartPosition(scimoz.currentPos, true);
    var end = scimoz.wordEndPosition(scimoz.currentPos, true);
    return scimoz.getTextRange(start, end);
}

function Interpolate_CurrentFilePath() {
    var editor = windowManager_getMainWindow();
    var view = editor.gViewMgr.currentView;
    if (view && view.document && view.document.file && view.document.file.isLocal) {
        return view.document.file.path;
    }
    return '';
}

function Interpolate_CurrentFileProjectPath() {
    // return the path of the project that the current file is in
    var editor = windowManager_getMainWindow();
    var view = editor.gViewMgr.currentView;
    var projectFile = '';
    if (view && view.document && view.document.file && view.document.file.isLocal) {
        var fileURL = view.document.file.URI;
        var filePart = editor.gProjectManager.findPartByTypeAttributeValue('file',
                                    'url', fileURL, true)
        if (filePart) {
            var projectPart = filePart.project;
            projectFile = uriparse_displayPath(projectPart.url);
        }
    } else {
        // use current project if there are no files open
        var project = gProjectManager.getCurrentProject();
        if (project) {
            projectFile = uriparse_displayPath(project.url);
        }
    }
    return projectFile;
}


function Interpolate_activeProjectPath() {
    // return the path of the active project
    var editor = windowManager_getMainWindow();
    var project = editor.gProjectManager.getCurrentProject();
    var projectFile = '';
    if (project) {
        projectFile = uriparse_displayPath(project.url);
    }
    return projectFile;
}
// A utility function to retrieving specific view data required by
// Interpolate_interpolate().
//  "editor" is a reference to the komodo.xul window.
//  "viewData" (optional) is a object possibly containing overriding values for
//      some of the view data.
function Interpolate_getViewData(editor,
                                 viewData /* =<determined from curr view> */)
{
    if (typeof viewData == 'undefined' || viewData == null)
        viewData = new Object();

    var view = editor.gViewMgr.currentView;
    var scimoz = null;
    if (view && view.getAttribute("type") == "editor") {
        scimoz = view.scintilla.scimoz;
    }

    // fileName
    if ( !("fileName" in viewData) ) {
        if (view) {
            viewData.fileName = view.document.displayPath;
        } else {
            viewData.fileName = null;
        }
    }

    // projectFile
    if ( !("projectFile" in viewData) ) {
        viewData.projectFile = Interpolate_activeProjectPath();
    }

    // lineNum
    if ( !("lineNum" in viewData) ) {
        if (scimoz) {
            viewData.lineNum = scimoz.lineFromPosition(scimoz.currentPos) + 1;
        } else {
            viewData.lineNum = 0;
        }
    }

    // selection
    if ( !("selection" in viewData) ) {
        if (scimoz) {
            viewData.selection = scimoz.selText;
        } else {
            viewData.selection = null;
        }
    }

    // word
    if ( !("word" in viewData) ) {
        if (scimoz) {
            viewData.word = Interpolate_getWordUnderCursor(scimoz);
        } else {
            viewData.word = null;
        }
    }

    // prefSet
    if ( !("prefSet" in viewData) ) {
        if (view) {
            viewData.prefSet = view.prefs;
        } else {
            viewData.prefSet = null;  // means: fallback to the global prefset
        }
    }

    return viewData;
}


// Interpolate '%'-escape codes in the given list(s) of strings.
//
//  "editor" is a reference the komodo.xul window.
//  "strings" is a list of raw strings to interpolate.
//  "bracketedStrings" is a list of raw strings to interpolate, using the bracketed form
//  "queryTitle" (optional) is a title for the possible query dialog raised
//      during interpolation.
//  "viewData" (optional) allows one to override specific view data used for
//      interpolation. By default view data is retrieved from the current view.
//      This may not always be appropriate. It may be an object with one or
//      more of the following attributes:
//          "fileName" is the filename of the current file (null, if N/A);
//          "lineNum" is the current line number (0, if N/A);
//          "word" is the current word under cursor (null if none);
//          "selection" is the current selection (null if none).
//
// On success, this function returns a *double* list of interpolated strings:
// For each string in "strings" and "bracketedStrings" two strings are
// returned. The first is the interpolated string for use and the second for
// *display*. In most cases these are the same but they may differ, for
// example, if a password query response was interpolated into the string which
// should not be displayed in the Komodo UI.
//
// Otherwise an exception is raised and an error set on the last error service:
//      koILastError errno      reason
//      ----------------------- -----------------------------------------
//      NS_ERROR_ABORT          User cancelled the query dialog.
//      NS_ERROR_INVALID_ARG    A normal interpolation failure because of
//                              invalid interp code usage.
//
function Interpolate_interpolate(editor, strings, bracketedStrings,
                                 queryTitle /* =null */,
                                 viewData /* =<determined from current view> */)
{
    try {
    if (typeof queryTitle == 'undefined') queryTitle = null;
    _interpolate_log.info("Interpolate_interpolate(editor, strings=["+strings+
                          "], bracketedStrings=["+bracketedStrings+"], queryTitle='"+
                          queryTitle+"', viewData)");
    viewData = Interpolate_getViewData(editor, viewData);

    var lastErrorSvc = Components.classes["@activestate.com/koLastErrorService;1"]
                       .getService(Components.interfaces.koILastErrorService);

    // Interpolation step 1: get queries.
    var queriesCountObj = new Object();
    var queriesObj = new Array();
    var i1countObj = new Object();
    var i1stringsObj = new Array();

    //XXX The prefset used may need to be the project prefs at some point in
    //    the future, for now we'll stick with the current view's prefset.
    var iSvc = Components.classes["@activestate.com/koInterpolationService;1"]
               .getService(Components.interfaces.koIInterpolationService);
    iSvc.Interpolate1(strings.length, strings,
                      bracketedStrings.length, bracketedStrings,
                      viewData.fileName, viewData.lineNum,
                      viewData.word, viewData.selection,
                      viewData.projectFile,
                      viewData.prefSet,
                      queriesCountObj, queriesObj,
                      i1countObj, i1stringsObj);
    var queries = queriesObj.value;
    var istrings = i1stringsObj.value;

    // Ask the user for required data. (If there are no queries then we are
    // done interpolating.)
    if (queries.length != 0) {
        var obj = new Object();
        obj.queries = queries;
        obj.title = queryTitle;
        window.openDialog("chrome://komodo/content/run/interpolationquery.xul",
                          "komodo_interpolationquery",
                          "chrome,modal,titlebar",
                          obj);
        if (obj.retval == "Cancel") {
            var errmsg = "Interpolation query cancelled.";
            lastErrorSvc.setLastError(Components.results.NS_ERROR_ABORT,
                                      errmsg);
            throw errmsg;
        }

        // Interpolation step 2: interpolated with answered queries.
        var i2countObj = new Object();
        var i2stringsObj = new Array();
        iSvc.Interpolate2(istrings.length, istrings,
                          queries.length, queries,
                          i2countObj, i2stringsObj);
        istrings = i2stringsObj.value;
    }

    _interpolate_log.info("Interpolate_interpolate: istrings=["+istrings+"]");
    return istrings;
    } catch(e) {
        _interpolate_log.exception(e);
        throw e;
    }
}


// Return a list of start and end offsets of interpolation code blocks into the
// given strings "s".
function Interpolate_getBlockOffsets(s, bracketed)
{
    var offsetsCountObj = new Object();
    var offsetsObj = new Array();
    var iSvc = Components.classes["@activestate.com/koInterpolationService;1"]
               .getService(Components.interfaces.koIInterpolationService);
    iSvc.GetBlockOffsets(s, bracketed, offsetsCountObj, offsetsObj);
    var offsets = offsetsObj.value;

    return offsets;
}


